from supermutes.register import get_new_obj, register
from supermutes.base import SuperMutable


dotify = lambda obj: get_new_obj(obj)
"Helper function for converting a mutable"


def reset_mapping():
    register(__name__, dict, DotDict)
    register(__name__, list, DotList)


class DotList(SuperMutable, list):
    """
    A list that allows dot notation access to its items.

    For example:

    >> d = DotList(['a', 'b', 'c'])
    >> d._0
    'a'
    >> d._1
    'b'
    >> d._2
    'c'
    >> d._3
    IndexError

    """
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        for i, value in enumerate(self):
            self[i] = value

    def mutate(self, value):
        return get_new_obj(value)

    def __getattr__(self, attr):
        try:
            index = int(attr.strip('_'))
            return self[index]
        except ValueError:
            pass
        return list.__getattr__(self, attr)

    def __setattr__(self, attr, value):
        try:
            index = int(attr.strip('_'))
            self[index] = value
        except ValueError:
            pass
        super(DotList, self).__setattr__(attr, value)


class DotDict(SuperMutable, dict):
    """
    A dictionary that allows dot notation access to its values.

    For example:

    >>  d = DotDict({
            'a': 12,
            'b': {
                'something': 5
            }
        })
    >>  d.a
    12
    >>  d.b.something
    5
    >>  d.c = 99
    >>  del d['a']
    >>  del d['b']
    >>  d
    {'c': 99}
    """
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        for key, value in self.items():
            self[key] = value

    def mutate(self, value):
        return get_new_obj(value)

    __getattr__ = dict.__getitem__
    __setattr__ = SuperMutable.__setitem__



from collections import defaultdict
import inspect

CLASS_REGISTER = defaultdict(dict)


def get_new_obj(obj):
    frm = inspect.stack()[1]
    module = inspect.getmodule(frm[0]).__name__
    register = CLASS_REGISTER[module]
    for old_class, new_class in register.items():
        if isinstance(obj, old_class) and not isinstance(obj, new_class):
            return new_class(obj)
    return obj


def register(module, old_class, new_class):
    CLASS_REGISTER[module][old_class] = new_class



from supermutes.register import get_new_obj
from supermutes.base import SuperMutable


class OnlyStringOrDict(SuperMutable, dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        for key, value in self.items():
            self[key] = value

    def mutate(self, value):
        return get_new_obj(value)  # get_new_obj uses the module that called
                                   # it to determine which kind of SuperMutable
                                   # is appropriate. This makes subclassing
                                   # of your SuperMutable easy!

    def __setitem__(self, key, value):
        if isinstance(value, basestring) or isinstance(value, dict):
            return super(OnlyStringOrDict, self).__setitem__(key, value)
        raise Exception("OnlyStringOrDict only likes strings or "
                        "dicts: {0}".format(value))

"""
In Use:

from supermutes.example import OnlyStringDict

>>  string_dict = OnlyStringOrDict({'one': 'f', 'two': {'what': '4'}})
>>  string_dict[4] = "This will go in fine"
>>  string_dict
{4: 'This will go in fine', 'two': {'what': '4'}, 'one': 'f'}
>>  string_dict[5] = {'another': "legitimate"}
>>  string_dict
{5: {'another': 'legitimate'}, 4: 'This will go in fine', 'two': {'what': '4'},
 'one': 'f'}
>>  string_dict[5]['fail'] = 10
Exception: OnlyStringOrDict only likes strings or dicts: 10
"""



import inspect

from register import register


def get_supermute_ancestor(cls):
    if SuperMutable in cls.__bases__:
        return cls
    for base in cls.__bases__:
        ancestor = get_supermute_ancestor(base)
        if ancestor:
            return ancestor


def get_mutable_type(cls):
    for mutable in [dict, list, set]:
        if mutable in inspect.getmro(cls):
            return mutable


class RegisterClass(type):
    """
    A metaclass used for registering any subclass as the main converter

    If it is the SuperMutable obj - don't worry too much it'll raise
    a NameError and that'll be all

    If it is an inherited class anywhere down from (and including)
    a supermute, then we need to find which kind of supermute it is so we can
    register against the appropriate module and mutable type.
    """
    def __init__(cls, name, bases, attrs):
        try:
            for base in bases:
                if issubclass(base, SuperMutable):
                    supermute = get_supermute_ancestor(cls)
                    map_to_class = get_mutable_type(supermute)
                    register(supermute.__module__, map_to_class, cls)
                    return
        except NameError:
            pass
        return


class SuperMutable(object):
    __metaclass__ = RegisterClass

    def mutate(self, value):
        raise NotImplementedError("Mutate() must be implemented on"
                                  " SuperMutable objects")

    def __setitem__(self, key, value):
        super(SuperMutable, self).__setitem__(key, self.mutate(value))



"A collection of readonly immutables"
from copy import deepcopy

from supermutes.register import get_new_obj, register
from supermutes.base import SuperMutable

readonly = lambda obj: get_new_obj(obj)


def reset_mapping():
    register(__name__, dict, ReadOnlyDict)
    register(__name__, list, ReadOnlyList)


class ReadOnlyClassException(Exception):
    pass


class ReadOnlyBaseClass(object):

    def method_not_allowed(self, *args, **kwargs):
        raise ReadOnlyClassException("Cannot write to object.")

    (__setitem__, __setattr__, __delitem__, __setslice__, __delslice__,
     append, insert, pop, popitem, remove, clear, update, extend,
     reverse, sort) = (method_not_allowed,) * 15


class ReadOnlyList(ReadOnlyBaseClass, SuperMutable, list):

    def __getitem__(self, index):
        return readonly(list.__getitem__(self, index))

    def get_writable(self):
        "Get a fully writable version of this list"
        writable = list([])
        for item in self:
            try:
                item = item.get_writable()
            except AttributeError:
                pass
            writable.append(item)
        return deepcopy(writable)


class ReadOnlyDict(ReadOnlyBaseClass, SuperMutable, dict):

    def __getitem__(self, key):
        return readonly(dict.__getitem__(self, key))

    def get_writable(self):
        "Get a fully writable version of this dictionary"
        writable = dict({})
        for key, item in self.items():
            try:
                item = item.get_writable()
            except AttributeError:
                pass
            writable[key] = item
        return deepcopy(writable)



#!/usr/bin/env python
# # coding: utf-8

from setuptools import setup
long_description = open('README').read()


setup(
    name='supermutes',
    description='A collection of super mutables',
    long_description=long_description,
    version='0.2.0',
    author='Alex Couper',
    author_email='info@alexcouper.com',
    url='https://github.com/alexcouper/supermutes',
    packages=['supermutes'],
    zip_safe=True,
    package_data={
        '': ['*.txt', '*.rst'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
